import{p as e,f as o}from"./flowDb-1972c806-9dd802e4.js";import{f as t,g as a}from"./styles-080da4f6-564d890e.js";import{t as s}from"./index-a5d3e69d.js";import"./layout-78cff630.js";import"./index-2c4b9a3b-4b03d70e.js";import"./edges-f2ad444c-22e77f4f.js";import"./createText-62fc7601-8b6fcc2a.js";import"./line-5038b469.js";import"./array-9f3ba611.js";import"./path-53f90ab3.js";const k={parser:e,db:o,renderer:t,styles:a,init:r=>{r.flowchart||(r.flowchart={}),r.flowchart.arrowMarkerAbsolute=r.arrowMarkerAbsolute,s({flowchart:{arrowMarkerAbsolute:r.arrowMarkerAbsolute}}),t.setConf(r.flowchart),o.clear(),o.setGen("gen-2")}};export{k as diagram};
